import os
import sys

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'SEPpy'
copyright = '2025, Jan Gieseler'
author = 'Jan Gieseler'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']




# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../'))



from astropy.utils.data import get_pkg_data_filename
from pathlib import Path
from seppy.tools import Event
from seppy.util import jupyterhub_data_path
import datetime
import os
import matplotlib.pyplot as plt
import pandas as pd
import pytest


"""
Install dependencies for tests:
pip install flake8 pytest pytest-doctestplus pytest-cov pytest-mpl

To create/update the baseline images, run the following command from the base package dir:
pytest --mpl-generate-path=seppy/tests/baseline seppy/tests/test_tools.py

To run the tests locally, go to the base directory of the repository and run:
pytest -ra --mpl --mpl-baseline-path=baseline --mpl-baseline-relative --mpl-generate-summary=html seppy/tests/test_tools.py
"""


# switch to non-plotting matplotlib backend to avoid showing all the figures:
plt.switch_backend("Agg")


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOLO_STEP_ions_old_data_online():
    startdate = datetime.date(2020, 9, 21)
    enddate = datetime.date(2020, 9, 21)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    background_range = (datetime.datetime(2020, 9, 21, 0, 0, 0), datetime.datetime(2020, 9, 21, 2, 0, 0))
    #
    # ions
    Event1 = Event(spacecraft='Solar Orbiter', sensor='STEP', viewing='Pixel averaged', data_level='l2', species='ions', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    # Pixel averaged
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='Pixel averaged', background_range=background_range, channels=1, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2020-09-21T17:27:37'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2020-09-21T17:57:37'
    assert fig.get_axes()[0].get_title() == 'SOLO/STEP 0.0060 - 0.0091 MeV/n protons\n5min averaging, viewing: PIXEL AVERAGED'
    # Pixel 8 - check that calculation is stopped bc. this data is not implemented correctly!
    check = False
    try:
        flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='Pixel 8', background_range=background_range, channels=1, resample_period="5min", yscale='log', cusum_window=30)
    except Warning:
        check = True
    assert check

    # TODO: deactivated, as this function is deactivated atm:
    # test dynamic spectrum:
    # Event1.dynamic_spectrum(view='Pixel averaged')
    # assert Event1.fig.get_axes()[0].get_title() == 'SOLO/STEP (Pixel averaged) ions, 2020-09-21'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('Pixel averaged', selection=(0, 4, 1), resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Solar Orbiter STEP, ions'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOLO_STEP_ions_new_data_online():
    startdate = datetime.date(2022, 1, 9)
    enddate = datetime.date(2022, 1, 9)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    background_range = (datetime.datetime(2022, 1, 9, 10, 0, 0), datetime.datetime(2022, 1, 9, 12, 0, 0))
    # ions
    Event1 = Event(spacecraft='Solar Orbiter', sensor='STEP', viewing='Pixel averaged', data_level='l2', species='ions', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    # Pixel averaged
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='Pixel averaged', background_range=background_range, channels=1, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (164,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:07:30'
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-01-09T01:32:31'
    assert fig.get_axes()[0].get_title() == 'SOLO/STEP 0.0061 - 0.0091 MeV protons\n5min averaging, viewing: PIXEL AVERAGED'
    # Pixel 8
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='Pixel 8', background_range=background_range, channels=1, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (164,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:12:30'
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-01-09T00:02:31'
    assert fig.get_axes()[0].get_title() == 'SOLO/STEP 0.0061 - 0.0091 MeV protons\n5min averaging, viewing: PIXEL 8'

    # TODO: deactivated, as this function is deactivated atm:
    # test dynamic spectrum:
    # Event1.dynamic_spectrum(view='Pixel averaged')
    # assert Event1.fig.get_axes()[0].get_title() == 'SOLO/STEP (Pixel averaged) ions, 2022-01-09'
    # Event1.dynamic_spectrum(view='Pixel 8')
    # assert Event1.fig.get_axes()[0].get_title() == 'SOLO/STEP (Pixel 8) ions, 2022-01-09'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('Pixel 8', selection=(0, 4, 1), resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Solar Orbiter STEP, ions'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOLO_HET_online():
    startdate = datetime.date(2022, 11, 8)
    enddate = datetime.date(2022, 11, 8)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    background_range = (datetime.datetime(2022, 11, 8, 0, 0, 0), datetime.datetime(2022, 11, 8, 1, 0, 0))
    # viewing "sun", single channel, protons
    Event1 = Event(spacecraft='Solar Orbiter', sensor='HET', viewing='sun', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='sun', background_range=background_range, channels=1, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (73,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5] == pd.Timestamp('2021-10-28 15:31:59.492059')
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-11-08T17:58:09'
    assert fig.get_axes()[0].get_title() == 'SOLO/HET 7.3540 - 7.8900 MeV protons\n5min averaging, viewing: SUN'
    # viewing "north", combined channel, electrons
    Event1 = Event(spacecraft='Solar Orbiter', sensor='HET', viewing='sun', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='north', background_range=background_range, channels=[0, 3], resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (73,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5] == pd.Timestamp('2021-10-28 15:31:59.492059')
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-11-08T22:27:56'
    assert fig.get_axes()[0].get_title() == 'SOLO/HET 0.4533 - 18.8300 MeV electrons\n5min averaging, viewing: NORTH'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='sun')
    assert Event1.fig.get_axes()[0].get_title() == 'SOLO/HET (sun) electrons, 2022-11-08'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('north', selection=None, resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Solar Orbiter HET, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOLO_EPT_online():
    startdate = datetime.date(2022, 6, 6)
    enddate = datetime.date(2022, 6, 6)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    background_range = (datetime.datetime(2022, 6, 6, 0, 0, 0), datetime.datetime(2022, 6, 6, 1, 0, 0))
    # viewing "sun", single channel, ions
    Event1 = Event(spacecraft='Solar Orbiter', sensor='EPT', viewing='sun', data_level='l2', species='ions', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='sun', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5] == pd.Timestamp('2021-10-28 15:31:59.492059')
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-06-06T01:02:31'
    assert fig.get_axes()[0].get_title() == 'SOLO/EPT 0.0608 - 0.0678 MeV protons\n5min averaging, viewing: SUN'
    # viewing "north", combined channel, electrons
    Event1 = Event(spacecraft='Solar Orbiter', sensor='EPT', viewing='sun', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='north', background_range=background_range, channels=[5, 10], resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)  # onset_stats[5] == pd.Timestamp('2021-10-28 15:31:59.492059')
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2022-06-06T01:32:31'
    assert fig.get_axes()[0].get_title() == 'SOLO/EPT 0.0439 - 0.0682 MeV electrons\n5min averaging, viewing: NORTH'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='sun')
    assert Event1.fig.get_axes()[0].get_title() == 'SOLO/EPT (sun) electrons, 2022-06-06'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('sun', selection=(0, 4, 1), resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Solar Orbiter EPT, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_PSP_ISOIS_EPIHI_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    # viewing "A", single channel, electrons
    Event1 = Event(spacecraft='PSP', sensor='isois-epihi', viewing='A', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='A', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (194,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:31:59'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T16:06:59'
    assert fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPIHI 0.8 - 1.0 MeV electrons\n5min averaging, viewing: A'
    # viewing "B", combined channel, protons
    Event1 = Event(spacecraft='PSP', sensor='isois-epihi', viewing='B', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='B', background_range=background_range, channels=[1, 5], resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (194,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:46:59'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T19:56:59'
    assert fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPIHI 8.0 - 19.0 MeV protons\n5min averaging, viewing: B'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='A')
    assert Event1.fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPIHI (A) protons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('A', selection=(0, 4, 1), resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Parker Solar Probe ISOIS-EPIHI, protons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_PSP_ISOIS_EPILO_e_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='PSP', sensor='isois-epilo', viewing='7', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    # viewing "7", single channel
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='7', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (198,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:33:14'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T16:43:14'
    assert fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPILO 65.9 - 100.5 keV electrons\n5min averaging, viewing: 7'
    # viewing "3", combined channels
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='3', background_range=background_range, channels=[0, 4], resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (198,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:33:14'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T17:48:14'
    assert fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPILO 10.0 - 100.5 keV electrons\n5min averaging, viewing: 3'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='7')
    assert Event1.fig.get_axes()[0].get_title() == 'PSP/ISOIS-EPILO (7) electrons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('3', selection=(0, 4, 1), resample='1min')
    assert plt.figure(1).get_axes()[0].get_title() == 'Parker Solar Probe ISOIS-EPILO, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_Wind_3DP_p_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='Wind', sensor='3DP', data_level='l2', viewing="Sector 3", species='protons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    # viewng "sector 3"
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='sector 3', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:27:35'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T20:17:35'
    assert fig.get_axes()[0].get_title() == 'WIND/3DP 385.96 - 716.78 keV protons\n5min averaging, viewing: SECTOR 3'
    # viewing "omnidirectional"
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='omnidirectional', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:27:42'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T21:07:42'
    assert fig.get_axes()[0].get_title() == 'WIND/3DP 385.96 - 716.78 keV protons\n5min averaging, viewing: OMNIDIRECTIONAL'
    # no channel combination inlcuded for Wind/3DP, yet

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='sector 3')
    assert Event1.fig.get_axes()[0].get_title() == 'WIND/3DP (sector 3) protons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('omnidirectional', selection=(0, 4, 1), resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'Wind 3DP, protons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_Wind_3DP_e_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='Wind', sensor='3DP', viewing="Sector 3", data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)  # TODO: radio_spacecraft=('wind', 'WIND')
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    #
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='sector 3', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:12:35'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T21:52:35'
    assert fig.get_axes()[0].get_title() == 'WIND/3DP 127.06 - 235.96 keV electrons\n5min averaging, viewing: SECTOR 3'
    #
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='omnidirectional', background_range=background_range, channels=4, resample_period="5min", yscale='log', cusum_window=30)
    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:07:42'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T21:52:42'
    assert fig.get_axes()[0].get_title() == 'WIND/3DP 127.06 - 235.96 keV electrons\n5min averaging, viewing: OMNIDIRECTIONAL'
    # no channel combination inlcuded for Wind/3DP, yet

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view='omnidirectional')
    assert Event1.fig.get_axes()[0].get_title() == 'WIND/3DP (omnidirectional) electrons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot('sector 3', selection=(0, 4, 1), resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'Wind 3DP, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_STEREOB_HET_p_online():
    startdate = datetime.date(2006, 12, 13)
    enddate = datetime.date(2006, 12, 14)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='STEREO-B', sensor='HET', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2006, 12, 13, 0, 0, 0), datetime.datetime(2006, 12, 13, 2, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing=None, background_range=background_range, channels=[5, 8], resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2006-12-13T03:18:04'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2006-12-13T09:53:04'
    assert fig.get_axes()[0].get_title() == 'STB/HET 26.3 - 40.5 MeV protons\n5min averaging'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view=None)
    assert Event1.fig.get_axes()[0].get_title() == 'STB/HET protons, 2006-12-13'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(None, selection=None, resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'STEREO-B HET, protons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_STEREOB_HET_e_online():
    startdate = datetime.date(2006, 12, 13)
    enddate = datetime.date(2006, 12, 14)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='STEREO-B', sensor='HET', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath, radio_spacecraft=('behind', 'STEREO-B'))
    print(Event1.print_energies())
    background_range = (datetime.datetime(2006, 12, 13, 0, 0, 0), datetime.datetime(2006, 12, 13, 2, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing=None, background_range=background_range, channels=[1], resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2006-12-13T02:43:04'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2006-12-13T04:53:04'
    assert fig.get_axes()[0].get_title() == 'STB/HET 1.4 - 2.8 MeV electrons\n5min averaging'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view=None)
    assert Event1.fig.get_axes()[0].get_title() == 'Radio & Dynamic Spectrum, STB/HET electrons, 2006-12-13'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(None, selection=None, resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'STEREO-B HET, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_STEREOA_SEPT_p_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 28)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='STEREO-A', sensor='SEPT', viewing="north", data_level='l2', species='ions', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='north', background_range=background_range, channels=[5, 8], resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:53:27'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T17:18:27'
    assert fig.get_axes()[0].get_title() == 'STA/SEPT 110-174.6 keV protons\n5min averaging, viewing: NORTH'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view="north")
    assert Event1.fig.get_axes()[0].get_title() == 'STA/SEPT (north) protons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(view="north", selection=None, resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'STEREO-A SEPT, protons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_STEREOA_SEPT_e_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 28)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='STEREO-A', sensor='SEPT', viewing="asun", data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath, radio_spacecraft=('ahead', 'STEREO-A'))
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='asun', background_range=background_range, channels=[8], resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:43:27'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T18:58:27'
    assert fig.get_axes()[0].get_title() == 'STA/SEPT 125-145 keV electrons\n5min averaging, viewing: ASUN'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view="asun")
    assert Event1.fig.get_axes()[0].get_title() == 'Radio & Dynamic Spectrum, STA/SEPT (asun) electrons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(view="asun", selection=None, resample=None)
    assert plt.figure(1).get_axes()[0].get_title() == 'STEREO-A SEPT, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOHO_EPHIN_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 28)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='SOHO', sensor='EPHIN', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing=None, background_range=background_range, channels=150, resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T15:53:42'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T22:18:42'  # '2021-10-29T04:53:42.357000')
    assert fig.get_axes()[0].get_title() == 'SOHO/EPHIN 0.25 - 0.7 MeV electrons\n5min averaging'
    # no channel combination inlcuded for SOHO/EPHIN electrons, yet

    # test dynamic spectrum:
    # check = False
    # try:
    #     Event1.dynamic_spectrum(view=None)
    # except Warning:
    #     check = True
    # assert check
    Event1.dynamic_spectrum(view=None)
    assert Event1.fig.get_axes()[0].get_title() == 'SOHO/EPHIN electrons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(None, selection=(0, 4, 1), resample='5min')
    assert plt.figure(1).get_axes()[0].get_title() == 'SOHO EPHIN, electrons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_spectrum_tsa_SOHO_ERNE_online():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    lpath = f"{os.getcwd()}{os.sep}data"
    lpath = jupyterhub_data_path(lpath)
    Event1 = Event(spacecraft='SOHO', sensor='ERNE-HED', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath, radio_spacecraft=('ahead', 'STEREO-A'))
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing=None, background_range=background_range, channels=3, resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:53:05'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T22:53:05'
    assert fig.get_axes()[0].get_title() == 'SOHO/ERNE 25.0 - 32.0 MeV protons\n5min averaging'

    # test dynamic spectrum:
    Event1.dynamic_spectrum(view=None)

    assert Event1.fig.get_axes()[0].get_title() == 'Radio & Dynamic Spectrum, SOHO/ERNE protons, 2021-10-28'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(None, selection=(0, 4, 1), resample='5min')
    assert plt.figure(1).get_axes()[0].get_title() == 'SOHO ERNE, protons'

    return fig


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_tsa_SOHO_ERNE_offline():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    fullpath = get_pkg_data_filename('data/test/soho_erne-hed_l2-1min_20211028_v01.cdf', package='seppy')
    lpath = Path(fullpath).parent.as_posix()
    Event1 = Event(spacecraft='SOHO', sensor='ERNE-HED', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath)
    print(Event1.print_energies())
    background_range = (datetime.datetime(2021, 10, 28, 10, 0, 0), datetime.datetime(2021, 10, 28, 12, 0, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing=None, background_range=background_range, channels=[1, 3], resample_period="5min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (288,)
    assert len(onset_stats) == 6
    assert onset_stats[5].isoformat().split('.')[0] == '2021-10-28T16:53:05'
    assert onset_found
    assert peak_time.isoformat().split('.')[0] == '2021-10-28T22:53:05'
    assert fig.get_axes()[0].get_title() == 'SOHO/ERNE 16.0 - 32.0 MeV protons\n5min averaging'

    # test tsa plot:
    plt.close('all')  # in order to pick the right figure, make sure all previous are closed
    Event1.tsa_plot(None, selection=(0, 4, 1), resample='5min')
    assert plt.figure(1).get_axes()[0].get_title() == 'SOHO ERNE, protons'

    return fig


def test_dynamic_spectrum_SOHO_ERNE_offline():
    startdate = datetime.date(2021, 10, 28)
    enddate = datetime.date(2021, 10, 29)
    fullpath = get_pkg_data_filename('data/test/soho_erne-hed_l2-1min_20211028_v01.cdf', package='seppy')
    lpath = Path(fullpath).parent.as_posix()
    radio_spacecraft = None  # use ('ahead', 'STEREO-A') if #27 is fixed and radio files can be provided offline
    Event1 = Event(spacecraft='SOHO', sensor='ERNE-HED', data_level='l2', species='protons', start_date=startdate, end_date=enddate, data_path=lpath, radio_spacecraft=radio_spacecraft)
    Event1.dynamic_spectrum(view=None)

    assert Event1.fig.get_axes()[0].get_title() == 'SOHO/ERNE protons, 2021-10-28'


@pytest.mark.mpl_image_compare(remove_text=False, deterministic=True)
def test_onset_Bepi_SIXS_offline():
    startdate = datetime.date(2023, 7, 19)
    enddate = datetime.date(2023, 7, 19)
    fullpath = get_pkg_data_filename('data/test/20230719_side1.csv', package='seppy')
    lpath = Path(fullpath).parent.as_posix()
    # lpath = '/home/jagies/data/bepi/bc_mpo_sixs/data_csv/cruise/sixs-p/raw'
    Event1 = Event(spacecraft='Bepi', sensor='SIXS', data_level='l2', species='electrons', start_date=startdate, end_date=enddate, data_path=lpath, viewing='1')
    background_range = (datetime.datetime(2023, 7, 19, 0, 30, 0), datetime.datetime(2023, 7, 19, 1, 30, 0))
    flux, onset_stats, onset_found, peak_flux, peak_time, fig, bg_mean = Event1.find_onset(viewing='1', background_range=background_range, channels=2, resample_period="1min", yscale='log', cusum_window=30)

    assert isinstance(flux, pd.Series)
    assert flux.shape == (161,)
    assert len(onset_stats) == 6
    assert isinstance(onset_stats[5], pd._libs.tslibs.nattype.NaTType)
    assert not onset_found
    assert peak_time.isoformat().split('.')[0] == '2023-07-19T02:25:30'
    assert fig.get_axes()[0].get_title() == 'BEPI/SIXS 106 keV electrons\n1min averaging, viewing: 1'

    return fig



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import cdflib
import copy
import datetime as dt
import numpy as np
import os
import pandas as pd
import pooch
import requests
import sunpy
import warnings

from sunpy.net import Fido
from sunpy.net import attrs as a

from seppy.util import custom_notification, custom_warning, resample_df


logger = pooch.get_logger()
logger.setLevel("WARNING")


def _download_metafile(dataset, path=None):
    """
    Download master cdf file from cdaweb for 'dataset'
    """
    if not path:
        path = sunpy.config.get('downloads', 'sample_dir')
    base_url = 'https://spdf.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/'
    fname = dataset.lower() + '_00000000_v01.cdf'
    url = base_url + fname
    try:
        downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=fname, path=path, progressbar=True)
    except ModuleNotFoundError:
        downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=fname, path=path, progressbar=False)
    #
    return downloaded_file


def wind3dp_download_fido(dataset, startdate, enddate, path=None, max_conn=5):
    """
    Downloads Wind/3DP CDF files via SunPy/Fido from CDAWeb

    Parameters
    ----------
    dataset : str
        Name of Wind/3DP dataset: \n
            - 'WI_SFSP_3DP': Electron omnidirectional fluxes 27 keV - 520 keV, often at 24 sec \n
            - 'WI_SFPD_3DP': Electron energy-angle distributions 27 keV to 520 keV, often at 24 sec \n
            - 'WI_SOSP_3DP': Proton omnidirectional fluxes 70 keV - 6.8 MeV, often at 24 sec \n
            - 'WI_SOPD_3DP': Proton energy-angle distributions 70 keV - 6.8 MeV, often at 24 sec \n
    startdate, enddate : datetime or str
        Datetime object (e.g., dt.date(2021,12,31) or dt.datetime(2021,4,15)) or
        "standard" datetime string (e.g., "2021/04/15") (enddate must always be
        later than startdate)
    path : str, optional
        Local path for storing downloaded data, by default None
    max_conn : int, optional
        The number of parallel download slots used by Fido.fetch, by default 5

    Returns
    -------
    List of downloaded files
    """
    trange = a.Time(startdate, enddate)
    cda_dataset = a.cdaweb.Dataset(dataset)
    try:
        result = Fido.search(trange, cda_dataset)
        filelist = [i[0].split('/')[-1] for i in result.show('URL')[0]]
        filelist.sort()
        if path is None:
            filelist = [sunpy.config.get('downloads', 'download_dir') + os.sep + file for file in filelist]
        elif type(path) is str:
            filelist = [path + os.sep + f for f in filelist]
        downloaded_files = filelist

        for i, f in enumerate(filelist):
            if os.path.exists(f) and os.path.getsize(f) == 0:
                os.remove(f)
            if not os.path.exists(f):
                downloaded_file = Fido.fetch(result[0][i], path=path, max_conn=max_conn)
        # downloaded_files = Fido.fetch(result, path=path, max_conn=max_conn)
        # downloaded_files.sort()
    except (RuntimeError, IndexError):
        print(f'Unable to obtain "{dataset}" data for {startdate}-{enddate}!')
        downloaded_files = []
    return downloaded_files


def wind3dp_single_download(file, path=None):
    """
    Download a single Wind/3DP level 2 data file from SRL Berkeley to local path

    Parameters
    ----------
    file : str
        file to be downloaded, e.g. 'wi_sfsp_3dp_20220602_v01.cdf'
    path : str
        local path where the files will be stored
    Returns
    -------
    downloaded_file : str
        full local path to downloaded file
    """

    # add a OS-specific '/' to end end of 'path'
    if path:
        if not path[-1] == os.sep:
            path = f'{path}{os.sep}'
    else:
        path = sunpy.config.get('downloads', 'download_dir') + os.sep

    data = file.split('_')[1]  # e.g. 'sfsp'
    year = file.split('_')[3][:4]
    base = f"https://sprg.ssl.berkeley.edu/wind3dp/data/wi/3dp/{data}/{year}/"

    url = base+'/'+file

    try:
        downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=file, path=path, progressbar=True)
    except ModuleNotFoundError:
        downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=file, path=path, progressbar=False)
    except requests.HTTPError:
        print(f'No corresponding data found at {url}')
        downloaded_file = []
    print('')

    return downloaded_file


def wind3dp_download(dataset, startdate, enddate, path=None, **kwargs):
    """
    Downloads Wind/3DP CDF files via SunPy/Fido from CDAWeb

    Parameters
    ----------
    dataset : {str}
        Name of Wind/3DP dataset: \n
            - 'WI_SFSP_3DP': Electron omnidirectional fluxes 27 keV - 520 keV, often at 24 sec \n
            - 'WI_SFPD_3DP': Electron energy-angle distributions 27 keV to 520 keV, often at 24 sec \n
            - 'WI_SOSP_3DP': Proton omnidirectional fluxes 70 keV - 6.8 MeV, often at 24 sec \n
            - 'WI_SOPD_3DP': Proton energy-angle distributions 70 keV - 6.8 MeV, often at 24 sec
    startdate, enddate : {datetime or str}
        Datetime object (e.g., dt.date(2021,12,31) or dt.datetime(2021,4,15)) or
        "standard" datetime string (e.g., "2021/04/15") (enddate must always be
        later than startdate)
    path : {str}, optional
        Local path for storing downloaded data, by default None

    Returns
    -------
    List of downloaded files
    """

    trange = a.Time(startdate, enddate)
    cda_dataset = a.cdaweb.Dataset(dataset)
    try:
        result = Fido.search(trange, cda_dataset)
        filelist = [i[0].split('/')[-1] for i in result.show('URL')[0]]
        filelist.sort()
        files = filelist
        if path is None:
            filelist = [sunpy.config.get('downloads', 'download_dir') + os.sep + file for file in filelist]
        elif type(path) is str:
            filelist = [path + os.sep + f for f in filelist]
        downloaded_files = filelist

        for i, f in enumerate(filelist):
            if os.path.exists(f) and os.path.getsize(f) == 0:
                os.remove(f)
            if not os.path.exists(f):
                # downloaded_file = Fido.fetch(result[0][i], path=path, max_conn=max_conn)
                downloaded_file = wind3dp_single_download(files[i], path=path)
                if downloaded_file == []:
                    print('Trying download from CDAWeb...')
                    downloaded_file = Fido.fetch(result[0][i], path=path)  #, max_conn=max_conn)

    except (RuntimeError, IndexError):
        print(f'Unable to obtain "{dataset}" data for {startdate}-{enddate}!')
        downloaded_files = []
    return downloaded_files


def _wind3dp_load(files, resample="1min", threshold=None):
    if isinstance(resample, str):
        try:
            _ = pd.Timedelta(resample)
        except ValueError:
            raise Warning(f"Your 'resample' option of [{resample}] doesn't seem to be a proper Pandas frequency!")
    # try:
    # read 0th cdf file
    # cdf = cdflib.CDF(files[0])
    # df = _cdf2df_3d(cdf, "Epoch")
    df =_read_cdf_wind3dp(files[0])

    # read additional cdf files
    if len(files) > 1:
        for f in files[1:]:
            # cdf = cdflib.CDF(f)
            # t_df = _cdf2df_3d(cdf, "Epoch")
            t_df =_read_cdf_wind3dp(f)
            df = pd.concat([df, t_df])

    # replace bad data with np.nan:
    df = df.replace(-np.inf, np.nan)

    # replace outlier data points above given threshold with np.nan
    # note: df.where(cond, np.nan) replaces all values where the cond is NOT fullfilled with np.nan
    # following Pandas Dataframe work is not too elegant, but works...
    if threshold:
        # create new dataframe of FLUX columns only with removed outliers
        df2 = df.filter(like='FLUX_').where(df.filter(like='FLUX_') <= threshold, np.nan)
        # drop these FLUX columns from original dataframe
        flux_cols = df.filter(like='FLUX_').columns
        df.drop(labels=flux_cols, axis=1, inplace=True)
        # add cleaned new FLUX columns to original dataframe
        df = pd.concat([df2, df], axis=1)

    if isinstance(resample, str):
        df = resample_df(df=df, resample=resample, pos_timestamp="center", origin="start")

    return df
    # except:
    #     raise Exception(f"Problem while loading CDF file! Delete downloaded file(s) {files} and try again. Sometimes this is enough to solve the problem.")


def wind3dp_load(dataset, startdate, enddate, resample="1min", multi_index=True,
                 path=None, threshold=None, **kwargs):
    """
    Load-in data for Wind/3DP instrument. Provides released data obtained by
    SunPy through CDF files from CDAWeb. Returns data as Pandas dataframe.

    Parameters
    ----------
    dataset : str
        Name of Wind/3DP dataset: \n
            - 'WI_SFSP_3DP': Electron omnidirectional fluxes 27 keV - 520 keV, often at 24 sec \n
            - 'WI_SFPD_3DP': Electron energy-angle distributions 27 keV to 520 keV, often at 24 sec \n
            - 'WI_SOSP_3DP': Proton omnidirectional fluxes 70 keV - 6.8 MeV, often at 24 sec \n
            - 'WI_SOPD_3DP': Proton energy-angle distributions 70 keV - 6.8 MeV, often at 24 sec \n
    startdate, enddate : datetime or str
        Datetime object (e.g., dt.date(2021,12,31) or dt.datetime(2021,4,15)) or
        "standard" datetime string (e.g., "2021/04/15") (enddate must always be
        later than startdate)
    resample : str, optional
        Frequency to which the original data (~24 seconds) is resamepled. Pandas
        frequency (e.g., '1min' or '1h') or None, by default "1min".
        Note that this is just a simple wrapper around thepandas
        resample function that is calculating the mean of the data in the new
        time bins. This is not necessarily the correct way to resample data,
        depending on the data type (for example for errors)!
    multi_index : bool, optional
        Provide output for pitch-angle resolved data as Pandas Dataframe with
        multiindex, by default True
    path : str, optional
        Local path for storing downloaded data, by default None
    threshold : int or float, optional
        Replace all FLUX values above 'threshold' with np.nan, by default None

    Returns
    -------
    _type_
        _description_
    """
    files = wind3dp_download(dataset, startdate, enddate, path)
    if len(files) > 0:
        df = _wind3dp_load(files, resample, threshold)

        # download master file from CDAWeb
        path_to_metafile = _download_metafile(dataset, path=path)

        # open master file from CDAWeb as cdf
        metacdf = cdflib.CDF(path_to_metafile)

        e_mean = df.filter(like='ENERGY_').mean()
        # from Table 3 of Wilson et al. 2021, https://doi.org/10.1029/2020RG000714
        # ∼30% ΔE/E => ΔE = 0.3*E
        if dataset in ['WI_SFSP_3DP', 'WI_SFPD_3DP', 'WI_SOSP_3DP', 'WI_SOPD_3DP']:
            delta_e = 0.3 * e_mean
        # ∼20% ΔE/E => ΔE = 0.2*E
        elif dataset in ['WI_ELSP_3DP', 'WI_ELPD_3DP', 'WI_EHSP_3DP', 'WI_EHPD_3DP']:
            delta_e = 0.2 * e_mean
        e_low = e_mean - delta_e
        e_high = e_mean + delta_e
        energies = pd.concat([e_mean, delta_e, 2*delta_e, e_low, e_high], axis=1, keys=['mean_E', 'delta_e', 'DE', 'lower_E', 'upper_E'])
        energies['Bins_Text']= np.around(e_low/1e3, 2).astype('string') +' - '+ np.around(e_high/1e3, 2).astype('string') + ' keV'

        for col in ['mean_E', 'delta_e', 'DE', 'lower_E', 'upper_E']:
            energies[col] = energies[col]/1e6  # convert energies to MeV

        custom_warning('The energy values in the metadata are now provided as MeV (previously eV)! Also DE is now 2*delta_e (previously just delta_e). Please adapt your scripts accordingly!')

        meta = {'channels_dict_df': energies,
                'ENERGY_UNITS': metacdf.varattsget('ENERGY')['UNITS'],
                'FLUX_UNITS': metacdf.varattsget('FLUX')['UNITS'],
                'FLUX_FILLVAL': metacdf.varattsget('FLUX')['FILLVAL'],
                }

        # for SFSP, SOSP, SFPD, SFSP:
        try:
            meta['APPROX_ENERGY_LABELS'] = metacdf.varget('APPROX_ENERGY_LABELS')
            meta['FLUX_LABELS'] = metacdf.varget('FLUX_ENERGY_LABL')
        except:
            pass

        # create multi-index data frame of flux
        if multi_index:
            if dataset == 'WI_SFPD_3DP' or dataset == 'WI_SOPD_3DP':
                no_channels = len(df[df.columns[df.columns.str.startswith("ENERGY")]].columns)
                t_df = [''] * no_channels
                multi_keys = np.append([f"FLUX_E{i}" for i in range(no_channels)],
                                       df.drop(df.columns[df.columns.str.startswith(f"FLUX_")], axis=1).columns,
                                       )
                for i in range(no_channels):
                    t_df[i] = df[df.columns[df.columns.str.startswith(f"FLUX_E{i}")]]
                t_df.extend([df[col] for col in df.drop(df.columns[df.columns.str.startswith(f"FLUX_")], axis=1).columns.values])
                df = pd.concat(t_df, axis=1, keys=multi_keys)
            else:
                print('')
                print('Multi-index function only available (and necessary) for pitch-angle resolved fluxes. Skipping.')
    else:
        df = []
        meta = ''
    return df, meta


wind_load = copy.copy(wind3dp_load)


"""
Modification of sunpy's read_cdf function to allow loading of 3-dimensional variables from a cdf file. Adjusted only for Wind/3DP cdf files. Skipping units.
This function is copied from sunpy under the terms of the BSD 2-Clause licence. See licenses/SUNPY_LICENSE.rst
"""


def _read_cdf_wind3dp(fname, ignore_vars=[]):
    """
    Read a CDF file that follows the ISTP/IACG guidelines.

    Parameters
    ----------
    fname : path-like
        Location of single CDF file to read.

    Returns
    -------
    list[GenericTimeSeries]
        A list of time series objects, one for each unique time index within
        the CDF file.

    References
    ----------
    Space Physics Guidelines for CDF https://spdf.gsfc.nasa.gov/sp_use_of_cdf.html
    """
    import astropy.units as u
    from cdflib.epochs import CDFepoch
    from packaging.version import Version
    from sunpy import log
    from sunpy.timeseries import GenericTimeSeries
    from sunpy.util.exceptions import warn_user
    cdf = cdflib.CDF(str(fname))
    # Extract the time varying variables
    cdf_info = cdf.cdf_info()
    meta = cdf.globalattsget()
    if hasattr(cdflib, "__version__") and Version(cdflib.__version__) >= Version("1.0.0"):
        all_var_keys = cdf_info.rVariables + cdf_info.zVariables
    else:
        all_var_keys = cdf_info['rVariables'] + cdf_info['zVariables']
    var_attrs = {key: cdf.varattsget(key) for key in all_var_keys}

    # Get keys that depend on time
    # var_keys = [var for var in var_attrs if 'DEPEND_0' in var_attrs[var] and var_attrs[var]['DEPEND_0'] is not None]
    # Manually define keys that depend on time for Wind/3DP cdf files, as they don't follow the standard
    var_keys = all_var_keys

    # Get unique time index keys
    # time_index_keys = sorted(set([var_attrs[var]['DEPEND_0'] for var in var_keys]))
    # Manually define time index key for Wind/3DP cdf files, as they don't follow the standard
    time_index_keys = [var_keys.pop(var_keys.index('Epoch'))]

    all_ts = []
    # For each time index, construct a GenericTimeSeries
    for index_key in time_index_keys:
        try:
            index = cdf.varget(index_key)
        except ValueError:
            # Empty index for cdflib >= 0.3.20
            continue
        # TODO: use to_astropy_time() instead here when we drop pandas in timeseries
        index = CDFepoch.to_datetime(index)
        df = pd.DataFrame(index=pd.DatetimeIndex(name=index_key, data=index))
        # units = {}

        # for var_key in sorted(var_keys):
        for var_key in var_keys:
            if var_key in ignore_vars:
                continue  # leave for-loop, skipping var_key

            attrs = var_attrs[var_key]
            # Skip the following check for Wind/3DP cdf files, as they don't follow the standard
            # # If this variable doesn't depend on this index, continue
            # if attrs['DEPEND_0'] != index_key:
            #     continue

            # Get data
            if hasattr(cdflib, "__version__") and Version(cdflib.__version__) >= Version("1.0.0"):
                var_last_rec = cdf.varinq(var_key).Last_Rec
            else:
                var_last_rec = cdf.varinq(var_key)['Last_Rec']
            if var_last_rec == -1:
                log.debug(f'Skipping {var_key} in {fname} as it has zero elements')
                continue

            data = cdf.varget(var_key)

            # Skip the following code block for Wind/3DP cdf files, as they don't follow the standard
            # # Set fillval values to NaN
            # # It would be nice to properley mask these values to work with
            # # non-floating point (ie. int) dtypes, but this is not possible with pandas
            # if np.issubdtype(data.dtype, np.floating):
            #     data[data == attrs['FILLVAL']] = np.nan

            # Skip the following code block for Wind/3DP cdf files, as they don't follow the standard
            # # Get units
            # if 'UNITS' in attrs:
            #     unit_str = attrs['UNITS']
            #     try:
            #         unit = u.Unit(unit_str)
            #     except ValueError:
            #         if unit_str in _known_units:
            #             unit = _known_units[unit_str]
            #         else:
            #             warn_user(f'astropy did not recognize units of "{unit_str}". '
            #                       'Assigning dimensionless units. '
            #                       'If you think this unit should not be dimensionless, '
            #                       'please raise an issue at https://github.com/sunpy/sunpy/issues')
            #             unit = u.dimensionless_unscaled
            # else:
            #     warn_user(f'No units provided for variable "{var_key}". '
            #               'Assigning dimensionless units.')
            #     unit = u.dimensionless_unscaled

            if data.ndim > 3:
                # Skip data with dimensions >= 3 and give user warning
                warn_user(f'The variable "{var_key}" has been skipped because it has more than 3 dimensions, which is unsupported.')
            elif data.ndim == 3:
                # Multiple columns, give each column a unique label.
                # Numbering hard-corded to Wind/3DP data!
                for j in range(data.T.shape[0]):
                    for i, col in enumerate(data.T[j, :, :]):
                        var_key_mod = var_key + f'_E{j}'
                        df[var_key_mod + f'_P{i}'] = col
                        # units[var_key_mod + f'_{i}'] = unit
            elif data.ndim == 2:
                # Multiple columns, give each column a unique label
                for i, col in enumerate(data.T):
                    df[var_key + f'_{i}'] = col
                    # units[var_key + f'_{i}'] = unit
            else:
                # Single column
                df[var_key] = data
                # units[var_key] = unit

        # all_ts.append(GenericTimeSeries(data=df, units=units, meta=meta))

    # if not len(all_ts):
    #     log.debug(f'No data found in file {fname}')
    return df  # all_ts[0].to_dataframe()




import datetime
import glob
import os
import warnings

import astropy.constants as const
import astropy.units as u
import numpy as np
import pandas as pd
import psutil
import sunpy.sun.constants as sconst
from sunpy.coordinates import get_horizons_coord
from tqdm.auto import tqdm

# Utilities toolbox, contains helpful functions


def custom_formatwarning(message, *args, **kwargs):
    """
    :meta private:
    """
    # ignore everything except the message
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = "\033[1m"
    return BOLD+RED+'WARNING: '+ENDC+ str(message) + '\n'


def custom_formatnotification(message, *args, **kwargs):
    """
    :meta private:
    """
    # ignore everything except the message
    YELLOW = '\033[93m'
    ENDC = '\033[0m'
    BOLD = "\033[1m"
    _yllw_str = '\x1b[38;5;226m'
    return BOLD+YELLOW+'NOTE: '+ENDC+ str(message) + '\n'


def custom_warning(message):
    """
    :meta private:
    """
    formatwarning_orig = warnings.formatwarning
    warnings.formatwarning = custom_formatwarning
    warnings.warn(message)
    warnings.formatwarning = formatwarning_orig
    return


def custom_notification(message):
    """
    :meta private:
    """
    formatwarning_orig = warnings.formatwarning
    warnings.formatwarning = custom_formatnotification
    warnings.warn(message)
    warnings.formatwarning = formatwarning_orig
    return


def resample_df(df, resample, pos_timestamp="center", origin="start"):
    """
    Resamples a Pandas Dataframe or Series to a new frequency. Note that this is
    just a simple wrapper around the pandas resample function that is
    calculating the mean of the data in the new time bins. This is not
    necessarily the correct way to resample data, depending on the data type
    (for example for errors)!

    Parameters
    ----------
    df : pd.DataFrame or pd.Series
            The dataframe or series to resample
    resample : str
            pandas-compatible time string, e.g., '1min', '2H' or '25s'
    pos_timestamp : str, default 'center'
            Controls if the timestamp is at the center of the time bin, or at
            the start of it
    origin : str, default 'start'
            Controls if the origin of resampling is at the first entry of the
            input dataframe/series (‘start’), or at the start of the day
            (‘start_day’)

    Returns
    -------
    df : pd.DataFrame or Series, depending on the input
    """
    # check if resample option makes sense (e.g., new frequency is smaller than original frequency)
    delta_resample = pd.to_timedelta(resample)
    delta_original = (df.index[-1] - df.index[-2]).floor('s')  # round to full seconds to avoid weirdness
    if delta_resample < delta_original:
        raise ValueError(f"Your resample option of '{resample}' is smaller than the original data cadence of '{delta_original}'. This is not supported!")
    elif delta_resample == delta_original:
        custom_warning(f"\nYour resample option of '{resample}' is equal to the original data cadence of '{delta_original}'. You should only average like this if you know EXACTLY what you are doing, as it could offset the position of the timestamps!\n")
    try:
        df = df.resample(resample, origin=origin, label="left").mean()
        if pos_timestamp == 'start':
            df.index = df.index
        else:
            df.index = df.index + pd.tseries.frequencies.to_offset(pd.Timedelta(resample)/2)
        # if pos_timestamp == 'stop' or pos_timestamp == 'end':
        #     df.index = df.index + pd.tseries.frequencies.to_offset(pd.Timedelta(resample))
    except ValueError:
        raise ValueError(f"Your resample option of '{resample} doesn't seem to be a proper Pandas frequency!")

    return df


def flux2series(flux, dates, cadence=None):
    """
    Converts an array of observed particle flux + timestamps into a pandas series
    with the desired cadence.

    Parameters
    ----------
    flux: an array of observed particle fluxes
    dates: an array of corresponding dates/times
    cadence: str - desired spacing between the series elements e.g. '1s' or '5min'

    Returns
    -------
    flux_series: Pandas Series object indexed by the resampled cadence
    """

    # from pandas.tseries.frequencies import to_offset

    # set up the series object
    flux_series = pd.Series(flux, index=dates)

    # if no cadence given, then just return the series with the original
    # time resolution
    if cadence is not None:
        flux_series = resample_df(df=flux_series, resample=cadence, pos_timestamp="center", origin="start")

    return flux_series


def bepicolombo_sixs_stack(path, date, side, pos_timestamp='center'):
    """
    :meta private:
    """
    # side is the index of the file here
    try:
        try:
            filename = f"{path}/sixs_phys_data_{date}_side{side}.csv"
            df = pd.read_csv(filename)
        except FileNotFoundError:
            # try alternative file name format
            filename = f"{path}/{date.strftime('%Y%m%d')}_side{side}.csv"
            df = pd.read_csv(filename)
            times = pd.to_datetime(df['TimeUTC'])
        # list comprehension because the method can't be applied onto the array "times"
        times = [t.tz_convert(None) for t in times]
        df.index = np.array(times)
        df = df.drop(columns=['TimeUTC'])

        if pos_timestamp == 'start':
            cadence1 = (df.index[1]-df.index[0]).seconds
            cadence2 = (df.index[-1]-df.index[-2]).seconds
            if cadence1 != cadence2:
                custom_warning("Bepi/SIXS cadence is changing throughout the day; something is wrong!")    
            df.index = df.index-pd.Timedelta(f'{cadence1/2}s')
        elif pos_timestamp == 'center':
            pass
    except FileNotFoundError:
        print(f'Unable to open {filename}')
        df = pd.DataFrame()
        filename = ''
    return df, filename


def bepi_sixs_load(startdate, enddate, side, path, pos_timestamp='center'):
    """
    :meta private:
    """
    dates = pd.date_range(startdate, enddate)

    # read files into Pandas dataframes:
    df, file = bepicolombo_sixs_stack(path, startdate, side=side, pos_timestamp=pos_timestamp)
    if len(dates) > 1:
        for date in dates[1:]:
            t_df, file = bepicolombo_sixs_stack(path, date.date(), side=side, pos_timestamp=pos_timestamp)
            df = pd.concat([df, t_df])

    channels_dict = {"Energy_Bin_str": {'E1': '71 keV', 'E2': '106 keV', 'E3': '169 keV', 'E4': '280 keV', 'E5': '960 keV', 'E6': '2240 keV', 'E7': '8170 keV',
                                        'P1': '1.1 MeV', 'P2': '1.2 MeV', 'P3': '1.5 MeV', 'P4': '2.3 MeV', 'P5': '4.0 MeV', 'P6': '8.0 MeV', 'P7': '15.0 MeV', 'P8': '25.1 MeV', 'P9': '37.3 MeV'},
                     "Electron_Bins_Low_Energy": np.array([55, 78, 134, 235, 1000, 1432, 4904]),
                     "Electron_Bins_High_Energy": np.array([92, 143, 214, 331, 1193, 3165, 10000]),
                     "Ion_Bins_Low_Energy": np.array([0.001, 1.088, 1.407, 2.139, 3.647, 7.533, 13.211, 22.606, 29.246]),
                     "Ion_Bins_High_Energy": np.array([1.254, 1.311, 1.608, 2.388, 4.241, 8.534, 15.515, 28.413, 40.0])}
    return df, channels_dict


def calc_av_en_flux_sixs(df, channel, species):
    """
    This function averages the flux of two energy channels of BepiColombo/SIXS-P into a combined energy channel
    channel numbers counted from 1

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing HET data
    channel : int or list
        energy channel or list with first and last channel to be used
    species : string
        'e', 'electrons', 'p', 'protons'

    Returns
    -------
    flux: pd.DataFrame
        channel-averaged flux
    en_channel_string: str
        string containing the energy information of combined channel

    :meta private:
    """

    # define constant geometric factors
    GEOMFACTOR_PROT8 = 5.97E-01
    GEOMFACTOR_PROT9 = 4.09E+00
    GEOMFACTOR_ELEC5 = 1.99E-02
    GEOMFACTOR_ELEC6 = 1.33E-01
    GEOMFACTOR_PROT_COMB89 = 3.34
    GEOMFACTOR_ELEC_COMB56 = 0.0972

    if type(channel) is list and len(channel) == 1:
        channel = channel[0]

    if species in ['p', 'protons']:
        if channel == [8, 9]:
            countrate = df['P8'] * GEOMFACTOR_PROT8 + df['P9'] * GEOMFACTOR_PROT9
            flux = countrate / GEOMFACTOR_PROT_COMB89
            en_channel_string = '37 MeV'
        else:
            print('No valid channel combination selected.')
            flux = pd.Series()
            en_channel_string = ''

    if species in ['e', 'electrons']:
        if channel == [5, 6]:
            countrate = df['E5'] * GEOMFACTOR_ELEC5 + df['E6'] * GEOMFACTOR_ELEC6
            flux = countrate / GEOMFACTOR_ELEC_COMB56
            en_channel_string = '1.4 MeV'
        else:
            print('No valid channel combination selected.')
            flux = pd.Series()
            en_channel_string = ''

    return flux, en_channel_string


"""
inf_inj_time.py
"""
SOLAR_ROT = sconst.get('sidereal rotation rate').to(u.rad/u.s)


def get_sun_coords(time='now'):
    '''
    Gets the astropy Sun coordinates.

    Args:
        time (datetime.datetime): time at which coordinates are fetched.

    Returns:
        sun coordinates.
    '''

    return get_horizons_coord("Sun", time=time)


def radial_distance_to_sun(spacecraft, time='now'):
    '''
    Gets the 3D radial distance of a spacecraft to the Sun.
    3D here means that it's the real spatial distance and not
    a projection on, say, the solar equatorial plane.

    Args:
        spacecraft (str): spacecraft to look for.
        time (datetime.datetime): time at which to look for.

    Returns:
        astropy units: radial distance.
    '''

    sc_coords = get_horizons_coord(spacecraft, time)

    return sc_coords.separation_3d(get_sun_coords(time=time))


def calc_spiral_length(radial_dist, sw_speed):
    '''
    Calculates the Parker spiral length from the Sun up to a given radial distance.

    Args:
        radial_dist (astropy units): radial distance to the Sun.
        sw_speed (astropy units): solar wind speed.

    Returns:
        astropy units: Parker spiral length.
    '''

    temp_const = ((SOLAR_ROT/sw_speed)*(radial_dist.to(u.km)-const.R_sun)).value
    sqrt_temp_const = np.sqrt(temp_const**2 + 1)

    return 0.5*u.rad * (sw_speed/SOLAR_ROT) * (temp_const*sqrt_temp_const + np.log(temp_const + sqrt_temp_const))


def calc_particle_speed(mass, kinetic_energy):
    '''
    Calculates the relativistic particle speed.

    Args:
        mass (astropy units): mass of the particle.
        kinetic_energy (astropy units): kinetic energy of the particle.

    Returns:
        astropy units: relativistic particle speed.
    '''

    gamma = np.sqrt(1 - (mass*const.c**2/(kinetic_energy + mass*const.c**2))**2)

    return gamma*const.c


def inf_inj_time(spacecraft, onset_time, species, kinetic_energy, sw_speed):
    '''
    Calculates the inferred injection time of a particle (electron or proton) from the Sun,
    given a detection time at some spacecraft.

    Args:
        spacecraft (str): name of the spacecraft.
        onset_time (datetime.datetime): time of onset/detection.
        species (str): particle species, 'p' or 'e'.
        kinetic_energy (astropy units): kinetic energy of particle. If no unit is supplied, is converted to MeV.
        sw_speed (astropy units): solar wind speed. If no unit is supplied, is converted to km/s.

    Returns:
        datetime.datetime: inferred injection time.
    '''

    if type(kinetic_energy) is not u.quantity.Quantity:
        kinetic_energy = kinetic_energy * u.MeV

    if type(sw_speed) is not u.quantity.Quantity:
        sw_speed = sw_speed * u.km/u.s

    mass_dict = {'p': const.m_p,
                 'e': const.m_e
                 }

    radial_distance = radial_distance_to_sun(spacecraft, time=onset_time)

    spiral_length = calc_spiral_length(radial_distance, sw_speed)
    particle_speed = calc_particle_speed(mass_dict[species], kinetic_energy)

    travel_time = spiral_length/particle_speed
    travel_time = travel_time.to(u.s)

    return onset_time - datetime.timedelta(seconds=travel_time.value), spiral_length.to(u.AU)


def energy2speed(species, kinetic_energy):
    '''
    Calculates the relativistic particle speed derived from kinetic energy.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        kinetic_energy (astropy units/float): kinetic energy of the particle using astropy units; if only float, MeV is used

    Returns:
        astropy units: relativistic particle speed.
    '''
    if type(kinetic_energy) is not u.quantity.Quantity:
        kinetic_energy = kinetic_energy * u.MeV
    mass_dict = {'p': const.m_p, 'e': const.m_e}
    return calc_particle_speed(mass_dict[species], kinetic_energy)


def speed2energy(species, speed):
    '''
    Calculates the relativistic kinetic energy derived from particle speed.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        speed (astropy units/float): relativistic particle speed of the particle using astropy units; if only float, m/s is used

    Returns:
        astropy units: kinetic energy
    '''
    if type(speed) is not u.quantity.Quantity:
        speed = speed * u.m/u.s
    mass_dict = {'p': const.m_p, 'e': const.m_e}
    gamma = 1/np.sqrt(1-speed**2/const.c**2)
    K = (gamma-1)*mass_dict[species]*const.c**2
    return K.to(u.MeV)


def speed2momentum(species, speed):
    '''
    Calculates the relativistic particle momentum derived from speed.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        speed (float/astropy units): kinetic energy of the particle using astropy units; if only float m/s is used

    Returns:
        astropy units: relativistic particle momentum
    '''
    if type(speed) is not u.quantity.Quantity:
        speed = speed * u.m/u.s
    mass_dict = {'p': const.m_p, 'e': const.m_e}
    return mass_dict[species] * speed


def energy2momentum(species, kinetic_energy):
    '''
    Calculates the relativistic particle momentum derived from kinetic energy.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        kinetic_energy (float/astropy units): kinetic energy of the particle using astropy units; if only float MeV is used

    Returns:
        astropy units: relativistic particle momentum
    '''
    if type(kinetic_energy) is not u.quantity.Quantity:
        kinetic_energy = kinetic_energy * u.MeV
    # mass_dict = {'p': const.m_p, 'e': const.m_e}
    return speed2momentum(species, energy2speed(species, kinetic_energy))


def intensity2psd(species, kinetic_energy, intensity):
    '''
    Calculates the phase space density (distribution function) derived from kinetic energy and intensity.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        kinetic_energy (astropy units): kinetic energy of the particle using astropy units
        intensity (astropy units): differential intensity using astropy units

    Returns:
        astropy units: phase space density

    Example:
        import astropy.units as u
        f = intensity2psd('e', 48*u.keV, 1e4/(u.cm**2 * u.sr * u.s * u.MeV))
    '''
    if type(kinetic_energy) is not u.quantity.Quantity or type(intensity) is not u.quantity.Quantity:
        print("All physical inputs have to be defined in astropy units! Run 'help(intensity2psd)' for an example.")
        return
    else:
        p = energy2momentum(species, kinetic_energy)
        f = intensity / (p**2)
        return f.to(u.s**3/(u.sr*u.kg**3*u.cm**6))


def intensity2vsd(species, kinetic_energy, intensity):
    '''
    Calculates the velocity space density (distribution function) derived from kinetic energy and intensity.

    Args:
        species (string): particle species (electrons 'e' or protons 'p') used to determine the particle mass
        kinetic_energy (astropy units): kinetic energy of the particle using astropy units
        intensity (astropy units): differential intensity using astropy units

    Returns:
        astropy units: phase space density

    Example:
        import astropy.units as u
        f = intensity2vsd('e', 48*u.keV, 1e4/(u.cm**2 * u.sr * u.s * u.MeV))
    '''
    if type(kinetic_energy) is not u.quantity.Quantity or type(intensity) is not u.quantity.Quantity:
        print("All physical inputs have to be defined in astropy units! Run 'help(intensity2psd)' for an example.")
        return
    else:
        mass_dict = {'p': const.m_p, 'e': const.m_e}
        p = energy2momentum(species, kinetic_energy)
        f = intensity / (p**2) * mass_dict[species]**3
        return f.to(u.s**3/(u.sr*u.cm**6))


def jupyterhub_data_path(path_org, path_hub='/home/jovyan/data'):
    """
    Checks if Notebook is run within SOLER/SERPENTINE JupyterHub. Returns the path of a common data folder if yes, and the original path if not.
    """
    if (psutil.Process().parent().name() == 'jupyterhub-sing') and all(x in psutil.Process().parent().environ()['PYTHONPATH'] for x in ['/home/jovyan/serpentine/', '/home/jovyan/soler/sep_tools/']):
        print(f"JupyterHub detected. Adjusting data path to {path_hub}")
        return path_hub
    else:
        return path_org


def remove_duplicate_cdf_files(path=None):
    """
    Removes duplicate .cdf files in the provided directory, keeping only the one with the highest version number.

    Parameters
    ----------
    path : string, optional
        Directory in which the .cdf files are. If None, current working directory will be used. By default None.

    Returns
    -------
    deleted_files : list
        List of deleted duplicate .cdf files.

    Examples
    --------
    >>> from seppy.util import remove_duplicate_cdf_files
    >>> deleted_files = remove_duplicate_cdf_files('/Users/johndoe/data/psp')
    Removing duplicate .cdf files in /Users/johndoe/data/psp/
    >>> print(deleted_files)
    []
    """
    if not path:
        path = os.getcwd()
    if path[-1] is not os.sep:
        path += os.sep
    all_cdf = glob.glob(f'{path}*.cdf')
    print(f'Removing duplicate .cdf files in {path}') 
    deleted_files = []
    for cdf in tqdm(all_cdf):
        cdf_wo_v = cdf.strip(cdf.split('_')[-1])
        cdf_duplicates = glob.glob(f'{cdf_wo_v}*')
        if len(cdf_duplicates) > 1:
            cdf_duplicates.sort(reverse=True)
            for dup in cdf_duplicates[1:]:
                deleted_files.append(dup)
                os.remove(dup)
    return deleted_files



import os

import cdflib
import pandas as pd
import pooch
import requests
import sunpy
from bs4 import BeautifulSoup
from packaging.version import Version
from seppy.util import resample_df
from sunpy.timeseries import TimeSeries

logger = pooch.get_logger()
logger.setLevel("WARNING")


def juice_radem_download(date, path=None):
    """Download JUICE/RADEM cruise science data file from ESA's PSA to local path

    Parameters
    ----------
    date : datetime object
        datetime of data to retrieve
    path : str
        local path where the files will be stored

    Returns
    -------
    downloaded_file : str
        full local path to downloaded file
    """
    # use sunpy download directory if no path is provided
    if not path:
        path = sunpy.config.get('downloads', 'download_dir')

    # add a OS-specific '/' to end end of 'path'
    if path:
        if not path[-1] == os.sep:
            path = f'{path}{os.sep}'

    # URL of the webpage containing the downloadable files
    base_url = f"https://archives.esac.esa.int/psa/ftp/Juice/juice_radem/data_raw/cruise/sc/{date.year}{date.strftime('%m')}/"

    # Send an HTTP GET request to the webpage
    response = requests.get(base_url)

    # Check if the request was successful
    if response.status_code == 200:
        soup = BeautifulSoup(response.content, 'html.parser')
        
        # Find all links on the page
        links = soup.find_all('a')

        # Filter for the file link 
        fname = None
        for link in links:
            href = link.get('href')
            if href and f"radem_raw_sc_{date.year}{date.strftime('%m')}{date.strftime('%d')}__" in href and href.endswith('.cdf'):
                fname = href
                break  # Get the first found link

        if fname:
            url = base_url + fname

            try:
                downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=fname, path=path, progressbar=True)
            except ModuleNotFoundError:
                downloaded_file = pooch.retrieve(url=url, known_hash=None, fname=fname, path=path, progressbar=False)
            except requests.HTTPError:
                print(f'No corresponding JUICE/RADEM data found at {url}')
                downloaded_file = []

            # # Download the file
            # file_response = requests.get(file_url)

            # # Save the file if the request was successful
            # if file_response.status_code == 200:
            #     fname = file_url.split('/')[-1]  # Extract fname from the URL
            #     with open(fname, 'wb') as f:
            #         f.write(file_response.content)
            #     print(f"Downloaded: {fname}")
            # else:
            #     print(f"Failed to download file: {file_response.status_code}")

            return downloaded_file
        else:
            print("No suitable file found online.")
            return None
    else:
        print(f"Failed to fetch the webpage: {response.status_code}")
        return None


def juice_radem_load(startdate, enddate, resample=None, path=None, pos_timestamp='center'):
    """Download & load JUICE/RADEM cruise science data and returns it as Pandas DataFrame (and metadata dictionaries).
    Note that the data is provided in counts and not converted to physical units (as of Nov 2025); also the instrument configuration changes over time.

    Parameters
    ----------
    startdate : datetime object
        start datetime of data to retrieve
    enddate : datetime object
        end datetime of data to retrieve
    resample : str
        resampling frequency (e.g. '1min', '10min', '1H', etc.). If None, no resampling is applied.
    path : str
        local path where the files are stored / will be downloaded to
    pos_timestamp : str
        position of the timestamp when resampling ('start', 'center', 'end')

    Returns
    -------
    df : Pandas DataFrame
        DataFrame containing the JUICE/RADEM data
    energies_dict : dict
        Dictionary containing the JUICE/RADEM data energy and label information
    metadata_dict : dict
        Dictionary containing the JUICE/RADEM data metadata
    """

    # Generate list of dates between startdate and enddate
    dates = pd.date_range(start=startdate, end=enddate, freq='D')

    downloaded_files = []
    for date in dates:
        fname = juice_radem_download(date, path=path)
        if fname:
            downloaded_files.append(fname)

    if not downloaded_files:
        print("No data files were downloaded.")
        return pd.DataFrame(), {}, {}

    # Load the data using SunPy TimeSeries
    data = TimeSeries(downloaded_files, concatenate=True)
    df = data.to_dataframe()

    # drop string columns
    df.drop(columns=['TIME_OBT'], inplace=True)

    # convert TIME_UTC column from string to datetime
    df['TIME_UTC'] = pd.to_datetime(df['TIME_UTC'])

    if resample:
        df = resample_df(df, resample, pos_timestamp=pos_timestamp)

    energies_dict, metadata_dict = juice_radem_load_metadata(filename=downloaded_files[0])

    return df, energies_dict, metadata_dict


def juice_radem_load_metadata(filename):
    """Load JUICE/RADEM cruise science data metadata and return it as a dictionary

    Returns
    -------
    energies_dict : dict
        Dictionary containing the JUICE/RADEM data energy and label information
    metadata_dict : dict
        Dictionary containing the JUICE/RADEM data metadata
    """

    # open cdf file with cdflib to access metadata
    cdf = cdflib.CDF(filename)

    # dict with all metadata info
    metadata_dict = {"Global_Attributes": cdf.globalattsget()}

    # dict with energy/label infos
    energies_dict = {}

    cdf_info = cdf.cdf_info()
    if hasattr(cdflib, "__version__") and Version(cdflib.__version__) >= Version("1.0.0"):
        all_var_keys = cdf_info.rVariables + cdf_info.zVariables
    else:
        all_var_keys = cdf_info['rVariables'] + cdf_info['zVariables']
    #
    for key in all_var_keys:
        metadata_dict[key] = cdf.varattsget(key)
        if cdf.varattsget(key)['VAR_TYPE'] == 'metadata':
            energies_dict[key] = cdf.varget(key)

    return energies_dict, metadata_dict
